include(ANSIC.mh)
include(version)
include(magic.m4)

/*
| kernel.mh
| C
|
*/ 
;;;;;;;;;;;;;;; i386 specific definitions for the LLL kernel ;;;;;;;;;;;;;;;;;;


errprint({C generic kernel.
})dnl


define({comment},{/*{$*}*/})

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Structures ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

{#include "config.h"}

struct API_Signature {
   int32 Magic_Number ;
   char  Object_Name[20] ;
   char  Version[8] ;
};

struc API_Header {
   int32 TotalLength ;		/* total length of the file */
   int32 EntryPoint ;		/* address of object to return */
   int32 Offset	;		/* offset where to load it */
   int32 HeaderSize ;		/* size of header fields in file */
   int32 ExpandedSize ;		/* size when expanded to memory */
   int32 ___ ;			/* fill in */
   int32 CheckSum ;		/* checksum */
   int32 LongCheckSum ;		/* long checksum */
})


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Some values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
define({TRUE},		(-1))	/* Logical True value */
define({FALSE},		0)	/* Logical False value */

BYTEBITS	=	 8  /* bits in a byte (smallest addressable value) */

/* CHARABITS	=	 0	; bits of char in address */
/* CELLABITS	=	 2	; bits of cell in address */
/* PAGEABITS	=	12	; bits of page in address */


;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Deduced values ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
CHARSIZE	=	(1 << CHARABITS)	; size of a char in bytes
CELLSIZE	=	(1 << CELLABITS)	; size of a cell in bytes
PAGESIZE	=	(1 << PAGEABITS)	; size of a page in bytes
CHARMASK	=	~ (CHARSIZE - 1)	; mask of chars in address
CELLMASK	=	~ (CELLSIZE - 1)	; mask of cells in address
PAGEMASK	=	~ (PAGESIZE - 1)	; mask of pages in address
BITSPERCHAR	=	CHARSIZE*BYTEBITS	; bits in a char
BITSPERCELL	=	CELLSIZE*BYTEBITS	; bits in a cell


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;; Minimal instruction set for a FORTH-like machine... ;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Registers:
; caller-saved: eax,ecx,edx
; callee-saved: ebx,esi,edi,ebp,esp
; constant:	cs,ds,es,ss,fs,gs
;  eax,edx	scratch; not saved upon a thread switch.
;  ecx		must be constant zero
;  ebx		buffers top of stack; saved upon a thread switch.
;  edi		data stack pointer (growing down)
;		If the stack is declared garbage-collection aware
;		all values on the stack may be corrected according to the
;		garbage collection on a "PAUSE", a "YIELD", a memory
;		allocation call or any such thing.
;		 And the thread will be unavailable during garbage
;		collection. Else, no relocatable address may be remembered
;		from a previous yield (), and none may be used at all if the
;		thread is preemptible.
;		Due to buffering, the stack must contain a ghost element
;		to be loaded into ebx when empty.
;  esp		local return stack pointer (growing down),
;		also heap pointer
;  ebp		frame on the heap
;  esi		points on current instruction to execute...
;
;
;
; Values:
; - de -1GB  +1GB-1
; - sinon, temporaire.
; pointeurs:
; - de +1GB  +3GB-1
; - sinon, non relogeable.


define({CELL},	CELLSIZE)		; a cell is 4 bytes long


;;; abstraction of the i386 instruction set

define({Add},		{($1) += ($2) ;})
define({Sub},		{($1) -= ($2) ;})
define({__Mov__},	{($1) = ($2) ;})
define({__Movb__},	{($1) = ($2) ;})
define({__Movw__},	{($1) = ($2) ;})
define({Neg},		{($1) = - ($1) ;})
define({Xchg},		{R0 = $1 ; $1 = $2 ; $2 = R0 ;})
define({},		{})
define({},		{})
define({},		{})
define({},		{})
define({},		{})

define({point},	{ifelse({$2},{},{*((cell*)$1)},{((cell*)$1)[$2]})})

define({PC},		{esi})	; program counter
define({DSP},		{edi})	; data stack pointer
define({DSINC},		{sizeof(cell)})	; increment for data stack -- grow up
define({TOS},		{ebx})	; buffered top of stack
define({RSP},		{esp})	; return stack pointer
define({RSINC},		{(-4)})	; increment for return stack -- grow down

define({FRAME},		{ebp})	; a frame pointer for heap manipulation
define({R0},		{eax})	; a scratch register
define({R0L},		{al})	; its lowest byte
define({R0H},		{ah})	; its second byte
define({R1},		{edx})	; second scratch register
define({COUNT},		{ecx})	; count register
define({COUNTLO},	{cl})	; count register -- low byte

define({INSTRTBL},	InstrT)		; table for bytecode threading

define({PopData},	{Mov({$1},TOS)
			 Add(DSP,-DSINC)})
define({PushData},	{Add(DSP,DSINC)
			 Mov(point(DSP),TOS)
			 Mov(TOS,{$1})})
define({PopReturn},	{ifelse({$1},,{Add(RSP,-RSINC)},{pop $1})})
define({PushReturn},	{push $1})
define({},		{})
define({},		{})
define({},		{})
define({},		{})

define({Enter},	{Add(PC,CELL)})
define({Next},	{jmp [PC]})


;;;;;;;;;;;;;;;;;;;;;;;;;; Defining FORTH words ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; defining a counted string
define({DefCString},
{__stringseg__({
locals({strlen},{strbeg},{strend},{
strbeg:
	.byte 	strlen
	defb($@)
strend:
strlen	=	strend-strbeg-1
})})})

;; defining a FORTH word


define({DefWord},{
ifelse({$2},,{$1},{$2}):
})





;;;;;;;;;;;;;;;;;;;;;;; Aerial (extra-light) threads ;;;;;;;;;;;;;;;;;;;;;;;;;
Struc({AT},{
; AerialThread	; These threads are lighter than light: Aerial !
   int32 SAVE_EBX	; initial Top of Stack
   int32 SAVE_ESI	; Data Stack Pointer
   int32 SAVE_EDI	; Return Stack Pointer
   int32 SAVE_EBP	; Those two are reserved for the
   int32 SAVE_ESP	;; threading mechanism.
   int32 SAVE_RESP	; ESP save when going ret-threaded
   int32 STATE		; flags used for thread sync'
   int32 MAX_ESI	; bottom of data stack
   int32 MAX_EDI	; bottom of return stack
   int32 LINKBACK	; Link Back to thread object
})



/*** Some instructions ***/

define({DefCompareWord0},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(TOS $4 0)?FALSE:TRUE ;
   Next
})dnl
define({DefCompareWord},
{DefWord({$1},{$2},{$3})
   Enter
   TOS=(NOS $4 TOS)?FALSE:TRUE ;
   Nip
   Next
})dnl
